from collections import defaultdict
import json
import random
import re
from os.path import expanduser
import time
from types import SimpleNamespace
from typing import List
from bossman.cache import cache
from bossman.errors import BossmanError, BossmanValidationError
from bossman.abc import ResourceTypeABC
from bossman.abc import ResourceABC
from bossman.abc import ResourceApplyResultABC
from bossman.config import ResourceTypeConfig
from bossman.plugins.akamai.cloudlet_v3 import CloudletAPIV3Client, CloudletAPIV3Error
from bossman.plugins.akamai.cloudlet_v3.data import Network, SharedPolicyActivationStatus, SharedPolicyAsCode, SharedPolicyVersion
from bossman.plugins.akamai.cloudlet_v3.error import PolicyNameNotFound, PolicyVersionValidationError
from bossman.plugins.akamai.cloudlet_v3.resource import SharedPolicyResource
from bossman.plugins.akamai.cloudlet_v3.schema import validate_policy_version
from bossman.plugins.akamai.cloudlet_v3.ui import SharedPolicyApplyResult, SharedPolicyStatus, SharedPolicyVersionStatus
from bossman.plugins.akamai.lib.edgegrid import EdgegridError
from bossman.plugins.akamai.utils import GenericVersionComments
from bossman.repo import Repo, Revision, RevisionDetails

RE_COMMIT = re.compile("^commit: ([a-z0-9]*)", re.MULTILINE)


class CloudletError(BossmanError):
  pass

class ResourceTypeOptions:
  def __init__(self, options):
    from os import environ
    self.env_prefix = options.get("env_prefix", "")
    self.edgerc = expanduser(environ.get("%sEDGERC" % self.env_prefix, options.get("edgerc", "~/.edgerc")))
    self.section = environ.get("%sEDGERC_SECTION" % self.env_prefix, options.get("section", "papi"))
    self.switch_key = environ.get("%sEDGERC_SWITCH_KEY" % self.env_prefix, options.get("switch_key", None))

class ResourceType(ResourceTypeABC):
  def __init__(self, repo: Repo, config: ResourceTypeConfig):
    super(ResourceType, self).__init__(repo, config)
    self.options = ResourceTypeOptions(config.options)
    self.client = CloudletAPIV3Client(self.options.edgerc, self.options.section, self.options.switch_key)

  def create_resource(self, path: str, **kwargs):
    return SharedPolicyResource(path, **kwargs)

  def get_resource_status(self, resource: SharedPolicyResource):
    try:
      policy = self.client.get_policy_by_name(resource.name)

      production_activations = dict()
      staging_activations = dict()
      interesting_versions = set()

      if policy is not None:
        if policy.currentActivations.production.latest != None:
          interesting_versions.add(policy.currentActivations.production.latest.policyVersion)
          production_activations[policy.currentActivations.production.latest.policyVersion] = policy.currentActivations.production.latest
        if policy.currentActivations.production.effective != None:        
          interesting_versions.add(policy.currentActivations.production.effective.policyVersion)
          production_activations[policy.currentActivations.production.effective.policyVersion] = policy.currentActivations.production.effective
        if policy.currentActivations.staging.latest != None:
          interesting_versions.add(policy.currentActivations.staging.latest.policyVersion)
          staging_activations[policy.currentActivations.staging.latest.policyVersion] = policy.currentActivations.staging.latest
        if policy.currentActivations.staging.effective != None:        
          interesting_versions.add(policy.currentActivations.staging.effective.policyVersion)
          staging_activations[policy.currentActivations.staging.effective.policyVersion] = policy.currentActivations.staging.effective
        latest_version = self.client.get_latest_policy_version(policy.id)
        latest_versions = []
        if len(interesting_versions):
          window = max(latest_version.version - min(interesting_versions) + 1, 10)
          latest_versions = self.client.get_latest_policy_versions(policy.id, window)
        elif latest_version != None:
          latest_versions = [latest_version]
        if latest_version != None:
          interesting_versions.add(latest_version.version)
      return SharedPolicyStatus(self.repo, resource, [
          SharedPolicyVersionStatus(
            self.repo,
            resource,
            v,
            production_activations.get(v.version, None),
            staging_activations.get(v.version, None)
          )
          for v in latest_versions
          if v.version in interesting_versions
        ],
        exists=True
      )
    except CloudletAPIV3Error as e:
      return SharedPolicyStatus(self.repo, resource, [], exists=False, error=e)

  def is_pending(self, resource: SharedPolicyResource, revision: Revision):
    notes = revision.get_notes(resource.path)
    if notes.get("has_errors", False):
      return False
    return not self.is_applied(resource, revision)

  def is_applied(self, resource: SharedPolicyResource, revision: Revision):
    notes = revision.get_notes(resource.path)
    policy_version = notes.get("policy_version", None)
    return policy_version is not None

  def get_revision_details(self, resource: SharedPolicyResource, revision: Revision) -> RevisionDetails:
    notes = revision.get_notes(resource.path)
    policy_version = notes.get("policy_version", None)
    return RevisionDetails(id=revision.id, details="v"+str(policy_version) if policy_version else None)

  def apply_change(self, resource: SharedPolicyResource, revision: Revision, previous_revision: Revision) -> ResourceApplyResultABC:
    # we should only call this function if the Revision concerns the resource
    assert len(set(resource.paths).intersection(revision.affected_paths)) > 0

    notes = SimpleNamespace(
      policy_id=None,
      policy_version=None,
      has_errors=False
    )

    try:

      try:
        # Get the policy json from the commit
        policy_data = revision.show_path(resource.policy_path, textconv=True)
        if policy_data is None:
          raise PolicyVersionValidationError("missing policy version JSON")
        policy_data = json.loads(policy_data)
        # before changing anything in API, check that the json files are valid
        validate_policy_version(policy_data)
        policy_data: SharedPolicyAsCode = self.client.converter.structure(policy_data, SharedPolicyAsCode)
      except PolicyVersionValidationError as e:
        notes.has_errors = True
        raise e

      try:
        # If the policy exists, retrieve it
        policy = self.client.get_policy_by_name(resource.name)
      except PolicyNameNotFound:
        # Otherwise, create it
        policy = self.client.create_policy(resource.name, policy_data.description, policy_data.cloudletType, policy_data.groupId)

      policy_version = self.client.create_policy_version(
        policy.id,
        # we have to limit the length of the description field to 255 characters due to limitations of the Cloudlets V3 API
        str(GenericVersionComments.from_revision(revision, 255)),
        policy_data.matchRules
      )

      notes.policy_id = policy.id
      notes.policy_version = policy_version.version

      return SharedPolicyApplyResult(resource, revision, policy_version)
    except RuntimeError as e:
        return SharedPolicyApplyResult(resource, revision, error=e)
    finally:
      # Finally, assign the updated values to the git notes
      revision.get_notes(resource.path).set(**vars(notes))

  def validate_working_tree(self, resource: SharedPolicyResource):
    try:
      with open(resource.policy_path, "r") as fd:
        policy_data = json.loads(fd.read())
        validate_policy_version(policy_data)
    except (NotADirectoryError, FileNotFoundError) as e:
      raise PolicyVersionValidationError("file not found {}".format(e.filename))

  def _release(self, network: Network, resource: SharedPolicyResource, revision: Revision, on_update: callable = lambda resource, status, progress: None):
    notes = revision.get_notes(resource.path)
    policy_id = notes.get("policy_id")
    if not policy_id:
      policy_id = self.client.get_policy_by_name(resource.name).id
    policy_version = notes.get("policy_version")
    if not policy_id:
      on_update(resource, "[magenta]Policy does not exist[/]", None)
      return
    if not policy_version:
      on_update(resource, "[magenta]Revision not deployed[/]", 1)
      return
    if notes.get("has_errors") == True:
      on_update(resource, ":boom: [grey53]v{:<3}[/] Policy version has validation errors".format(policy_version), 1)
      return

    describe = lambda activation_status: "[{}]{}[/] [grey53]v{:<3}[/] [bright_white]{}[/]".format(network.color, network.alias, policy_version, activation_status)

    try:
      on_update(resource, describe("STARTING"), None)
      activation = self.client.activate_policy_version(policy_id, policy_version, network)
      on_update(resource, describe(activation.status.value), 0.5)
      while activation.status == SharedPolicyActivationStatus.IN_PROGRESS:
        time.sleep(random.randint(-10, 10) + 20)
        activation = self.client.get_policy_version_activation_status(activation)
        on_update(resource, describe(activation.status.value), 0.5)
      on_update(resource, describe(activation.status.value), 1)
    except CloudletAPIV3Error as e:
      on_update(resource, ":boom: [grey53]v{policy_version:<3}[/] {error}".format(policy_version=policy_version, error=e), 1)
    except RuntimeError as e:
      on_update(resource, ":boom: [grey53]v{policy_version:<3}[/] An error occurred: {e}".format(policy_version=policy_version, e=e), 1)

  def prerelease(self, resource: ResourceABC, revision: Revision, on_update: callable = lambda resource, status, progress: None):
    return self._release(Network.STAGING, resource, revision, on_update)

  def release(self, resource: ResourceABC, revision: Revision, on_update: callable = lambda resource, status, progress: None):
    return self._release(Network.PRODUCTION, resource, revision, on_update)



from bossman.errors import BossmanError
from concurrent.futures import ThreadPoolExecutor, wait
from os import getcwd
import sys
import git
import argparse
from rich import get_console
from rich import print
from rich.padding import Padding
from rich.traceback import Traceback
from bossman import Bossman
from bossman.abc import ResourceABC

console = get_console()

def init(subparsers: argparse._SubParsersAction):
  parser = subparsers.add_parser("apply", help="apply local changes to remotes")
  parser.add_argument("-e", "--exact-match", action="store_true", default=False, help="match resource exactly")
  parser.add_argument("--force", action="store_true", default=False, help="don't skip dirty resources")
  parser.add_argument("--dry-run", action="store_true", default=False, help="show what would be applied, but don't actually do anything")
  parser.add_argument("--since", default=None, help="apply only revisions since this commit ref (useful to skip early history)")
  parser.add_argument("glob", nargs="*", default="*", help="select resources by glob pattern")
  parser.set_defaults(func=exec)

def exec(bossman: Bossman, glob, exact_match:bool, force:bool, dry_run:bool, since, **kwargs):
  resources = bossman.get_resources(*glob, exact_match=exact_match)
  if len(resources) == 0:
    print('no resources selected')
    return
  futures = []
  had_errors = False
  with ThreadPoolExecutor(10, "apply") as executor:
    for resource in resources:
      futures.append(executor.submit(apply_changes, bossman, resource, force, dry_run, since))
  for resource, future in zip(resources, futures):
    try:
      had_errors = future.result() or had_errors
    except Exception as e:
      had_errors = True
      print(":exclamation_mark:", resource)
      console.print_exception()
  print(":cookie: [green]all resources up to date[green]")
  if had_errors:
    print("[red]apply completed, but some errors occurred[/red]")
    sys.exit(3)

def apply_changes(bossman: Bossman, resource: ResourceABC, force: bool, dry_run: bool, since: str):
  try:
    status = bossman.get_resource_status(resource)
    revisions = bossman.get_missing_revisions(resource, since_rev=since)
    todo = len(revisions)
    had_errors = False
    if todo > 0:
      if status.dirty and not force:
        print(":stop_sign:", resource, "[magenta]dirty, skipping[/magenta]")
        return
      results = []
      for revision in revisions:
        try:
          results.append(bossman.apply_change(resource, revision, dry_run))
        except Exception as e:
          if not force:
            raise e
          results.append(":exclamation_mark: {} an error occurred while applying {}\n{}".format(resource, revision, e))
          console.print_exception()
      for result in results:
        had_errors = had_errors or result.had_errors
        print(result)
    else:
      print(":white_check_mark:", resource, "is up to date")
  except RuntimeError as e:
    had_errors = True
    print(":exclamation_mark:", resource, e)
  return had_errors



import pytest
from bossman.plugins.akamai.property import PropertyResource
from bossman import get_resources


@pytest.fixture(scope="module")
def resources():
    resources = list()
    resources.append(PropertyResource("dist/akamai/property/global.mydomain.com", name="global.mydomain.com"))
    resources.append(PropertyResource("dist/akamai/property/testglobal.mydomain.com", name="testglobal.mydomain.com"))
    resources.append(PropertyResource("dist/akamai/property/logistics.mydomain.com", name="logistics.mydomain.com"))
    resources.append(PropertyResource("dist/akamai/property/logistics.mydomain.com-related-domains", name="logistics.mydomain.com-related-domains"))
    return resources

@pytest.mark.parametrize(
    "glob, expected_result",
    (
        (["*"], ["global.mydomain.com", "testglobal.mydomain.com", "logistics.mydomain.com", "logistics.mydomain.com-related-domains"]),
        (["*/logistics.mydomain.com*"], ["logistics.mydomain.com", "logistics.mydomain.com-related-domains"]),
        (["*/logistics.mydomain.com*"], ["logistics.mydomain.com", "logistics.mydomain.com-related-domains"]),
        (["*/logistics.mydomain.com"], ["logistics.mydomain.com"]),
        (["dist/akamai/property/logistics.mydomain.com"], ["logistics.mydomain.com"]),
        (["*global.mydomain.com"], ["global.mydomain.com", "testglobal.mydomain.com"]),
        (["*/global.mydomain.com"], ["global.mydomain.com"]),
    ),
)

def test_glob_exact_match(resources, glob, expected_result):
    result = get_resources(resources, glob, True)
    assert(len(result) == len(expected_result))
    for r in result:
        assert(r.name in expected_result)



import pytest
from bossman.plugins.akamai.property import PropertyResource
from bossman import get_resources


@pytest.fixture(scope="module")
def resources():
    resources = list()
    resources.append(PropertyResource("dist/akamai/cloudlets/ER_bossman_demo", name="ER_bossman_demo"))
    resources.append(PropertyResource("dist/akamai/property/www-dev", name="www-dev"))
    resources.append(PropertyResource("dist/akamai/property/www-integration", name="www-integration"))
    resources.append(PropertyResource("dist/akamai/property/www-prod", name="www-prod"))
    return resources

@pytest.mark.parametrize(
    "glob, expected_result",
    (
        (["*"], ["ER_bossman_demo", "www-dev", "www-integration", "www-prod"]),
        (["www"], ["www-dev", "www-integration", "www-prod"]),
    ),
)

def test_glob_default(resources, glob, expected_result):
    result = get_resources(resources, glob, False)
    assert(len(result) == len(expected_result))
    for r in result:
        assert(r.name in expected_result)




from os import getenv
from posixpath import expanduser
import git
import json
from bossman.config import ResourceTypeConfig
from bossman.errors import BossmanError
from rich import print

config = ResourceTypeConfig({
  "module": "bossman.plugins.akamai.cloudlet_v3",
  "pattern": "akamai/cloudlet/{name}",
  "options": {
    "edgerc": getenv("AKAMAI_EDGERC", expanduser("~/.edgerc")),
    "section": getenv("AKAMAI_EDGERC_SECTION", "default"),
    "switch_key": getenv("AKAMAI_EDGERC_SWITCHKEY", None),
  }
})

POLICY_NAME = 'ER_ak_hogg_fr_shared_policy'

##############################################################################

import git
from bossman.plugins.akamai.cloudlet_v3.client import CloudletAPIV3Client

client = CloudletAPIV3Client(config.options['edgerc'], config.options['section'], config.options.get('switch_key', None))
policy = client.get_policy_by_name(POLICY_NAME)
print(policy)

##############################################################################

policy_version = client.get_policy_version(policy.id, 1)
print(policy_version)

##############################################################################

from bossman.plugins.akamai.cloudlet_v3.resourcetype import ResourceType

rt = ResourceType(git.Repo(), config)

res = rt.create_resource('akamai/cloudlet/' + POLICY_NAME, name=POLICY_NAME)
print(res.name)
print(rt.get_resource_status(res))


##############################################################################

from bossman.plugins.akamai.cloudlet_v3.schema import validate_policy_version

data = json.loads(
  """
  {
      "descrip22tion": "Initial version",
      "matchRules": [
          {
              "end": 0,
              "name": "Redirect images",
              "matchURL": "/images/*",
              "redirectURL": "/static/images/*",
              "start": 0,
              "statusCode": 302,
              "type": "erMatchRule",
              "useIncomingQueryString": true,
              "useRelativeUrl": "relative_url"
          }
      ]
  }
  """
)

try:
  validate_policy_version(data)
except BossmanError as e:
  print(e)


import re
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import List, Dict, Optional



class Network(Enum):
  PRODUCTION = 'PRODUCTION'
  STAGING = 'STAGING'
  
  @property
  def alias(self) -> str:
    return re.sub("[aeiou]", "", self.value, flags=re.IGNORECASE)[:3].upper()
  
  @property
  def color(self) -> str:
    return "green" if self == Network.PRODUCTION else "magenta"

class SharedPolicyActivationOperation(Enum):
  ACTIVATION = 'ACTIVATION'
  DEACTIVATION = 'DEACTIVATION'

class SharedPolicyActivationStatus(Enum):
  IN_PROGRESS = 'IN_PROGRESS'
  SUCCESS = 'SUCCESS'
  FAILED = 'FAILED'

class SharedPolicyPolicyType(Enum):
  SHARED = 'SHARED'

class SharedPolicyCloudletType(Enum):
  ER = 'ER'
  FR = 'FR'
  AS = 'AS'
  CD = 'CD'
  IG = 'IG'

@dataclass
class SharedPolicyActivation:
  id: int
  createdBy: str
  createdDate: datetime
  finishDate: Optional[datetime]
  network: Network
  operation: SharedPolicyActivationOperation
  status: SharedPolicyActivationStatus
  policyId: int
  policyVersion: int
  policyVersionDeleted: bool

@dataclass
class SharedPolicyCurrentActivationsNetwork:
  effective: Optional[SharedPolicyActivation] = None
  latest: Optional[SharedPolicyActivation] = None

@dataclass
class SharedPolicyCurrentActivations:
  production: SharedPolicyCurrentActivationsNetwork = None
  staging: SharedPolicyCurrentActivationsNetwork = None

@dataclass
class SharedPolicy:
  id: int
  name: str
  description: str
  policyType: SharedPolicyPolicyType
  cloudletType: SharedPolicyCloudletType
  createdBy: str
  createdDate: datetime
  currentActivations: SharedPolicyCurrentActivations
  groupId: int
  modifiedBy: str
  modifiedDate: datetime

@dataclass
class GetPoliciesResponse:
  content: List[SharedPolicy]

@dataclass
class SharedPolicyVersion:
  policyId: int
  version: int
  createdBy: str
  createdDate: datetime
  modifiedBy: str
  modifiedDate: datetime
  description: str
  matchRules: Optional[List[Dict]] = None

@dataclass
class GetPolicyVersionsResponse:
  content: List[SharedPolicyVersion]

@dataclass
class SharedPolicyAsCode:
  """
  Subset of SharedPolicy that must be versioned in order to support CRUD.
  """
  description: str
  groupId: int
  cloudletType: SharedPolicyCloudletType
  matchRules: Optional[List[Dict]] = None



import jsonschema, jsonschema.validators, json
import pkg_resources

from bossman.plugins.akamai.cloudlet_v3.data import SharedPolicyAsCode

from .error import *

_PACKAGE = __name__.rsplit('.', 1)[0]

def get_schema(name):
  path = pkg_resources.resource_filename(_PACKAGE, 'schemas/' + name)
  with open(path) as fd:
    return json.loads(fd.read())

def validate_policy_version(data: dict):
  try:
    schema = get_schema('update-policy-version.json')
    validator = jsonschema.validators.Draft4Validator(schema, resolver=jsonschema.validators.RefResolver(
      pkg_resources.resource_filename(_PACKAGE, 'schemas/'),
      pkg_resources.resource_filename(_PACKAGE, 'schemas/update-policy-version.json'),
    ))
    for f in ['match_rule-AS-1.0.json', 'match_rule-ER-1.0.json', 'match_rule-FR-1.0.json', 'match-rules.json']:
      validator.resolver.store['file://' + f] = get_schema(f)
    validator.validate(data)
  except json.decoder.JSONDecodeError as err:
    raise PolicyVersionValidationError("bad rules.json", err.args)
  except jsonschema.ValidationError as err:
    raise PolicyVersionValidationError("invalid rules.json", {
      "message": err.message,
      "path": '/' + '/'.join(str(i) for i in err.absolute_path)
    })



from bossman.plugins.akamai.lib.edgegrid import EdgegridError
import json
from os.path import expanduser
from types import SimpleNamespace

from bossman.errors import BossmanError, BossmanValidationError
from bossman.abc import ResourceTypeABC
from bossman.abc import ResourceStatusABC
from bossman.abc import ResourceABC
from bossman.abc import ResourceApplyResultABC
from bossman.plugins.akamai.utils import GenericVersionComments
from bossman.repo import Repo, Revision, RevisionDetails
from bossman.plugins.akamai.lib.papi import (
  PAPIClient,
  PAPIEdgeHostnameChangeAlreadyInProgressError,
  PAPIEdgeHostnameNotAvailableError,
  PAPIError,
)
from bossman.rich import bracketize
from rich import print

class EdgeHostnameValidationError(BossmanValidationError):
  pass

class EdgeHostnameNotFoundError(BossmanError):
  pass

class EdgeHostnameResource(ResourceABC):
  def __init__(self, path, **kwargs):
    super(EdgeHostnameResource, self).__init__(path)
    self.__name = kwargs.get("name")

  @property
  def certEnrollmentId(self):
    return self.__name

  @property
  def name(self):
    return self.__name

  @property
  def paths(self):
    return (self.path,)

  def __rich__(self):
    path = self.path.replace(self.name, f'[/][yellow]{self.name}[/][grey53]')
    return f'[grey53]{path}[/]'

class EdgeHostnameStatus(ResourceStatusABC):
  def __init__(self, repo: Repo, resource: EdgeHostnameResource, error=None):
    self.repo = repo
    self.resource = resource
    self.error = error

  @property
  def exists(self):
    return False

  @property
  def dirty(self) -> bool:
    if not self.exists:
      return False
    return True

  def __rich_console__(self, *args, **kwargs):
    yield "[gray31](status not implemented for EdgeHostnames yet)[/]"

class EdgeHostnameApplyResult(ResourceApplyResultABC):
  def __init__(self,
              resource: EdgeHostnameResource,
              revision: Revision,
              error=None):
    self.resource = resource
    self.revision = revision
    self.error = error

  @property
  def had_errors(self) -> bool:
    return self.error != None

  def __rich_console__(self, *args, **kwargs):
    parts = []
    parts.append(r':arrow_up:')
    parts.append(self.resource.__rich__())
    parts.append(r'[grey53]{h}[/]'.format(h=bracketize(self.revision.id)))
    yield " ".join(parts)
    if self.error is not None:
      from rich.syntax import Syntax
      import yaml
      error_yaml = yaml.safe_dump(self.error.args[0])
      yield '{}\n{}'.format(type(self.error).__name__, Syntax(error_yaml, "yaml").highlight(error_yaml))

class ResourceTypeOptions:
  def __init__(self, options):
    from os import environ
    self.env_prefix = options.get("env_prefix", "")
    self.edgerc = expanduser(environ.get("%sEDGERC" % self.env_prefix, options.get("edgerc", "~/.edgerc")))
    self.section = environ.get("%sEDGERC_SECTION" % self.env_prefix, options.get("section", "default"))
    switch_key_opt = options.get("account_key", options.get("switch_key", None))
    if "switch_key" in options:
      print('[red]WARNING: "switch_key" option is deprecated; use "account_key" instead[/]')
    self.switch_key = environ.get("%sEDGERC_SWITCH_KEY" % self.env_prefix, switch_key_opt)

class ResourceType(ResourceTypeABC):
  def __init__(self, repo: Repo, config):
    super(ResourceType, self).__init__(repo, config)
    self.options = ResourceTypeOptions(config.options)
    self.papi = PAPIClient(self.options.edgerc, self.options.section, self.options.switch_key)

  def create_resource(self, path: str, **kwargs):
    return EdgeHostnameResource(path, **kwargs)

  def get_resource_status(self, resource: EdgeHostnameResource):
    try:
      return EdgeHostnameStatus(self.repo, resource, [])
    except BossmanError as e:
      return EdgeHostnameStatus(self.repo, resource, [], error=e)

  def is_pending(self, resource: EdgeHostnameResource, revision: Revision):
    return not self.is_applied(resource, revision)

  def is_applied(self, resource: EdgeHostnameResource, revision: Revision):
    if self.affects(resource, revision):
      notes = revision.get_notes(resource.path)
      return notes.get('is_applied', False)
    return True

  def get_revision_details(self, resource: ResourceABC, revision: Revision) -> RevisionDetails:
    return RevisionDetails(id=revision.id, details=None)

  def apply_change(self, resource: EdgeHostnameResource, revision: Revision, previous_revision: Revision=None) -> EdgeHostnameApplyResult:
    # we should only call this function if the Revision concerns the resource
    assert self.affects(resource, revision)

    notes = SimpleNamespace(
      is_applied=None,
      has_errors=False,
      edgehostname_id=None,
    )

    try:
      data = revision.show_path(resource.path, textconv=True)
      args = self.validate(resource, data)
      result = self.papi.create_edgehostname(**args)
      notes.edgehostname_id = result.edgeHostnameId
      notes.is_applied = True
      return EdgeHostnameApplyResult(resource, revision, error=None)
    except (PAPIEdgeHostnameNotAvailableError, PAPIEdgeHostnameChangeAlreadyInProgressError) as e:
      notes.is_applied = True
      notes.has_errors = False
      return EdgeHostnameApplyResult(resource, revision)
    except (EdgeHostnameValidationError, PAPIError) as e:
      notes.is_applied = True
      notes.has_errors = False
      return EdgeHostnameApplyResult(resource, revision, error=e)
    except EdgegridError as e:
      return EdgeHostnameApplyResult(resource, revision, error=e)
    finally:
      revision.get_notes(resource.path).set(**vars(notes))

  def validate_working_tree(self, resource: EdgeHostnameResource):
    try:
      with open(resource.path, "r") as hfd:
        self.validate(resource, hfd.read())
    except (NotADirectoryError, FileNotFoundError) as e:
      raise EdgeHostnameValidationError("file not found {}".format(e.filename))

  def validate(self, resource, data):
    if data is None:
      raise EdgeHostnameValidationError(f'{resource.path}: missing')
    try:
      data = json.loads(data)
    except json.JSONDecodeError as e:
      raise EdgeHostnameValidationError(f'{resource.path}: ({e.lineno}:{e.colno}) {e.msg}')
    requiredFields = ("groupId", "contractId", "edgeHostname")
    missing = list(field for field in requiredFields if field not in data)
    if not "edgeHostname" in missing:
      requiredFields = ("domainPrefix", "domainSuffix", "ipVersionBehavior", "productId", "secureNetwork")
      missing += list(f'edgeHostname.{field}' for field in requiredFields if field not in data["edgeHostname"])
      if "secureNetwork" in data["edgeHostname"] and data["edgeHostname"]["secureNetwork"] == "ENHANCED_TLS":
        missing += list(f'edgeHostname.{field}' for field in ["certEnrollmentId"] if field not in data["edgeHostname"])
    if len(missing):
      raise EdgeHostnameValidationError(f'{resource.path}: missing required fields: {", ".join(missing)}')
    return dict(
      groupId=data.get('groupId'),
      contractId=data.get('contractId'),
      **data.get('edgeHostname'),
    )

  def prerelease(self, resource: ResourceABC, revision: Revision, on_update: callable = lambda resource, status, progress: None):
    on_update(resource, '[grey53]STG Nothing to do[/]', 1)

  def release(self, resource: ResourceABC, revision: Revision, on_update: callable = lambda resource, status, progress: None):
    on_update(resource, '[grey53]PRD Nothing to do[/]', 1)

  def get_ehn_id(self, ehn: EdgeHostnameResource):
    revision = self.repo.get_last_revision(paths=[ehn.path])
    if revision == None:
      self.get_working_copy_ehn_data(ehn)
      return None
    blob = revision.show_path(ehn.path, textconv=True)
    data = json.loads(blob)
    groupId = data.get('groupId')

  def get_working_copy_ehn_data(self, ehn: EdgeHostnameResource):
    with open(ehn.path, 'r') as fd:
      print(fd.read())



from os import getcwd
import git
import argparse
from rich import print, box
from rich.table import Table
from bossman import Bossman
from bossman.abc import ResourceStatusABC

def init(subparsers: argparse._SubParsersAction):
  parser = subparsers.add_parser("status", help="show resource status")
  parser.add_argument("-e", "--exact-match", action="store_true", default=False, help="match resource exactly")
  parser.add_argument("glob", nargs="*", default="*", help="select resources by glob pattern")
  parser.set_defaults(func=exec)

def exec(bossman: Bossman, glob, exact_match:bool, *args, **kwargs):
  resources = bossman.get_resources(*glob, exact_match=exact_match)
  if len(resources) == 0:
    print('no resources selected')
    return
  if len(resources):
    print("On branch", bossman.get_current_branch())

    statuses = bossman.get_resource_statuses(resources)

    col_width = min(max(len(resource.path) for resource in resources) + 2, 60)
    table = Table(expand=False, box=box.HORIZONTALS, show_lines=True)
    table.add_column("Resource", width=col_width)
    table.add_column("Status")
    for resource, status in zip(resources, statuses):
      table.add_row(resource, status)
    print(table)
  else:
    print("No resources to show: check the glob pattern if provided, or the configuration.")



import platform
from os.path import abspath, commonpath, join
from fnmatch import fnmatch
from functools import cached_property

from rich.console import Console
from rich.progress import Progress

from bossman.errors import BossmanError
from bossman.resources import ResourceManager
from bossman.abc import ResourceApplyResultABC, ResourceStatusABC
from bossman.abc import ResourceTypeABC
from bossman.abc import ResourceABC
from bossman.config import Config, ResourceTypeConfig
from bossman.logging import get_class_logger
from bossman.repo import Repo, Revision, Change

import pkg_resources
DISTRIBUTION = pkg_resources.require("bossman")[0]
USER_AGENT = f'{DISTRIBUTION.project_name.capitalize()}/{DISTRIBUTION.version} (Python/{platform.python_version()}; {platform.system()}/{platform.release()}; +https://github.com/ynohat/bossman)'

def if_initialized(func):
  def wrapper(bossman, *args, **kwargs):
    if bossman.initialized:
      return func(bossman, *args, **kwargs)
    raise BossmanError("Repository not initialized or needs migration, please run `bossman init`.")
  return wrapper

def get_resources(resources, globs, exact=False) -> list:
  if exact:
    # exact match on filename
    globs = list(globs)
  else:
    # enable partial matches
    globs = list("*" + glob.strip("*") + "*" for glob in globs)
  # by default, all resources
  if len(globs) == 0:
    globs = globs.append("*")
  match = lambda resource: any(fnmatch(resource.path, glob) for glob in globs)
  return list(sorted(filter(match, resources)))

class DryRunApplyResult(ResourceApplyResultABC):
  def __init__(self, resource: ResourceABC, revision: Revision):
    self.resource = resource
    self.revision = revision

  @property
  def had_errors(self) -> bool:
    """
    Return True if any errors occurred applying a change to the resource.
    """
    return False

  def __rich_console__(self, *args, **kwargs):
    parts = []
    parts.append(r'[dark_orange3](--dry-run)[/]')
    parts.append(r':arrow_up:')
    parts.append(self.resource.__rich__())
    parts.append(r'[grey53][{h}][/]'.format(h=self.revision.id))
    if self.revision.short_message:
      parts.append(r'[bright_white]"{subject_line}"[/]'.format(subject_line=self.revision.short_message))
    author = self.revision.author_name
    if author:
      parts.append("[grey53]{}[/]".format(author))
    yield " ".join(parts)

class Bossman:
  def __init__(self, root, config: Config):
    self.config = config
    self.root = root
    self.logger = get_class_logger(self)

  @cached_property
  def repo(self):
    return Repo(self.root)

  @cached_property
  def resource_manager(self):
    return ResourceManager(self.config.resource_types, self.repo)

  @cached_property
  def initialized(self):
    config = self.repo.config_reader()
    if "bossman" in config.sections():
      from .migrations import required
      if required(config):
        return False
      return True
    return False

  @cached_property
  def version(self):
    import pkg_resources
    return pkg_resources.require("bossman")[0].version

  def init(self, console: Console):
    self.init_git_config(console)

  def init_git_config(self, console: Console):
    from .migrations import migrate
    conf = self.repo.config_writer()
    migrate(conf, self.repo, console)
    conf.set_value("bossman", "version", self.version)


  @if_initialized
  def get_resources(self, *globs, rev: str = "HEAD", exact_match: bool = False) -> list:
    resources = self.resource_manager.get_resources(self.repo, rev)
    return get_resources(resources, list(globs), exact_match)

  @if_initialized
  def get_resources_from_working_copy(self, *globs, exact_match: bool = False) -> list:
    resources = self.resource_manager.get_resources_from_working_copy(self.repo)
    return get_resources(resources, list(globs), exact_match)

  @if_initialized
  def is_applied(self, resource: ResourceABC, rev: Revision) -> bool:
    resource_type = self.resource_manager.get_resource_type(resource.path)
    return resource_type.is_applied(resource, rev)

  @if_initialized
  def get_missing_revisions(self, resource: ResourceABC, since_rev: str = None) -> list:
    resource_type = self.resource_manager.get_resource_type(resource.path)
    revisions = self.get_revisions(resources=[resource], since_rev=since_rev)
    missing = []
    for revision in revisions:
      # The revision may affect a file that is in the resource folder, but
      # is not managed by the resource. Only add the revision if it actually
      # affects it.
      if resource_type.affects(resource, revision):
        if resource_type.is_pending(resource, revision):
          missing.append(revision)
        else:
          break
    return list(reversed(missing))

  @if_initialized
  def get_revision_details(self, resource: ResourceABC, revision: Revision):
    resource_type = self.resource_manager.get_resource_type(resource.path)
    return resource_type.get_revision_details(resource, revision)

  @if_initialized
  def get_resource_statuses(self, resources: list) -> list:
    from concurrent.futures import ThreadPoolExecutor
    futures = []
    with ThreadPoolExecutor(10, "get_status") as executor:
      for resource in resources:
        futures.append(executor.submit(self.get_resource_status, resource))
    return [f.result() for f in futures]

  @if_initialized
  def get_resource_status(self, resource: ResourceABC) -> ResourceStatusABC:
    self.repo.fetch_notes(resource.path)
    resource_type = self.resource_manager.get_resource_type(resource.path)
    return resource_type.get_resource_status(resource)

  @if_initialized
  def get_revision(self, rev: str = None, resources: list = None) -> Revision:
    return self.repo.get_revision(rev, (p for r in resources for p in r.paths))

  @if_initialized
  def get_head_revision(self) -> Revision:
    return self.get_revision()

  @if_initialized
  def get_current_branch(self) -> str:
    return self.repo.get_current_branch()

  @if_initialized
  def get_revisions(self, since_rev: str = None, until_rev: str = "HEAD", resources: list = None) -> str:
    return self.repo.get_revisions(since_rev, until_rev, resources)

  @if_initialized
  def apply_change(self, resource: ResourceABC, revision: Revision, dry_run: bool):
    if dry_run:
      return DryRunApplyResult(resource, revision)
    self.repo.fetch_notes(resource.path)
    previous_revision = self.repo.get_last_revision(resource.paths, revision.parent_id)
    resource_type = self.resource_manager.get_resource_type(resource.path)
    result = resource_type.apply_change(resource, revision, previous_revision)
    self.repo.push_notes(resource.path)
    return result

  @if_initialized
  def validate(self, resource: ResourceABC):
    resource_type = self.resource_manager.get_resource_type(resource.path)
    resource_type.validate_working_tree(resource)

  @if_initialized
  def prerelease(self, resource: ResourceABC, revision: Revision, on_update: callable):
    self.repo.fetch_notes(resource.path)
    resource_type = self.resource_manager.get_resource_type(resource.path)
    resource_type.prerelease(resource, revision, on_update)
    self.repo.push_notes(resource.path)

  @if_initialized
  def release(self, resource: ResourceABC, revision: Revision, on_update: callable):
    self.repo.fetch_notes(resource.path)
    resource_type = self.resource_manager.get_resource_type(resource.path)
    resource_type.release(resource, revision, on_update)
    self.repo.push_notes(resource.path)

  # @if_initialized
  # def prerelease(self, resources: list, revision: Revision):
  #   resource_types = set(self.resource_manager.get_resource_type(resource.path) for resource in resources)
  #   for resource_type in resource_types:
  #     _resources = resource_type.get_resources(list(resource.path for resource in resources))
  #     resource_type.prerelease(_resources, revision)

  # @if_initialized
  # def release(self, resources: list, revision: Revision):
  #   resource_types = set(self.resource_manager.get_resource_type(resource.path) for resource in resources)
  #   for resource_type in resource_types:
  #     _resources = resource_type.get_resources(list(resource.path for resource in resources))
  #     resource_type.release(_resources, revision)



from abc import ABC, abstractmethod, abstractproperty
import sys
import git, gitdb
from os.path import basename
from bossman.errors import BossmanError
from bossman.logging import get_class_logger
from datetime import datetime
import gitdb.exc
import yaml
import threading

from bossman.rich import bracketize

def true(*args, **kwargs):
  return True

def short_rev(rev):
  return rev[:8] if rev else rev

class RepoError(BossmanError):
  pass
class RepoAuthError(RepoError):
  pass
class RepoRevNotFoundError(RepoError):
  pass

class TreeVisitor:
  """
  Callable to visit all the blobs in a tree.
  """
  def __init__(self, func):
    self.func = func

  def __call__(self, tree: git.objects.tree.Tree):
    def visit(blob, visited):
      if blob.path in visited:
        return # avoid infinite recursion
      visited.add(blob.path)
      if type(blob) == git.Tree:
        for child in blob:
          visit(child, visited)
      elif type(blob) == git.Blob:
        self.func(blob)
    visit(tree, set())

class Change(ABC):
  @staticmethod
  def from_diff(diff: git.diff.Diff):
    if diff.change_type == "R":
      return RenameChange(diff)
    if diff.change_type == "A":
      return AddChange(diff)
    if diff.change_type == "M":
      return ModifyChange(diff)
    if diff.change_type == "D":
      return DeleteChange(diff)
    raise RepoError("unknown change type {}".format(diff.change_type))

  def __init__(self, diff: git.Diff):
    self.diff = diff

  @property
  def path(self):
    return self.diff.b_path

  @property
  def basename(self):
    return basename(self.path)

  @property
  def change_type(self):
    return self.diff.change_type

  @property
  def contents(self):
    return self.diff.b_blob.data_stream.read()

  @property
  def diff_text(self):
    return str(self.diff)

  def __str__(self):
    return "{} {}".format(self.diff.change_type, self.path)

  def __rich_console__(self, console, options):
    color = "yellow"
    yield "[{color}]{change_type} {name}[{color}]".format(
      color=color,
      change_type=self.diff.change_type,
      name=self.basename,
    )

class AddChange(Change):
  pass

class RenameChange(Change):
  @property
  def from_path(self):
    return self.diff.a_path

class ModifyChange(Change):
  pass

class DeleteChange(Change):
  @property
  def path(self):
    return self.diff.a_path

  @property
  def contents(self):
    return self.diff.a_blob.data_stream.read()

class Notes:
  def __init__(self, repo, commit: git.Commit, ns: str = None):
    self.repo = repo
    self.commit = commit
    self.ns = ns.strip("/")

  def set(self, **kwargs):
    notes = self.read()
    notes.update(kwargs)
    self.write(notes)

  def get(self, k: str, default=None):
    notes = self.read()
    return notes.get(k, default)

  def read(self) -> dict:
    with self.repo._lock:
      try:
        args = []
        if self.ns is not None:
          args.extend(("--ref", self.ns))
        args.extend(("show", self.commit.hexsha))
        notes = self.commit.repo.git.notes(*args)
        return yaml.safe_load(notes if notes else '{}')
      except git.GitCommandError:
        return {}

  def write(self, notes):
    try:
      with self.repo._lock:
        args = []
        if self.ns is not None:
          args.extend(("--ref", self.ns))
        args.extend(("add", "-f", "-m", yaml.safe_dump(notes), self.commit.hexsha))
        self.commit.repo.git.notes(*args)
    except git.GitCommandError as err:
      raise BossmanError(err)



class Revision:
  def __init__(self, repo, commit: git.Commit, diffs: git.Diff):
    self.repo = repo
    self.commit = commit
    self.changes = dict(
      (diff.b_path or diff.a_path, Change.from_diff(diff))
      for diff
      in diffs
    )

  @property
  def id(self):
    with self.repo._lock:
      return short_rev(self.commit.hexsha)

  @property
  def parent_id(self):
    try:
      with self.repo._lock:
        return short_rev(self.commit.parents[0].hexsha)
    except IndexError:
      return None

  @property
  def branches(self):
    with self.repo._lock:
      return self.repo.get_branches_containing(self.id)

  @property
  def date(self) -> datetime:
    with self.repo._lock:
      return self.commit.committed_datetime

  @property
  def message(self) -> str:
    with self.repo._lock:
      return self.commit.message

  @property
  def short_message(self) -> str:
    with self.repo._lock:
      return self.message.split("\n").pop(0)

  @property
  def author_name(self) -> str:
    with self.repo._lock:
      return self.commit.author.name

  @property
  def author_email(self) -> str:
    with self.repo._lock:
      return self.commit.author.email

  @property
  def committer_name(self) -> str:
    with self.repo._lock:
      return self.commit.committer.name

  @property
  def committer_email(self) -> str:
    with self.repo._lock:
      return self.commit.committer.email

  @property
  def affected_paths(self) -> list:
    with self.repo._lock:
      return self.changes.keys()

  def show_path(self, path, textconv=False) -> dict:
    with self.repo._lock:
      contents = self.show_paths([path], textconv)
      return contents.get(path, None)

  def show_paths(self, paths: list, textconv=False) -> dict:
    with self.repo._lock:
      contents = dict()
      def get_blob_contents(blob):
        if blob.path in paths:
          if textconv:
            contents[blob.path] = self.repo._repo.git.show(
              '--textconv',
              '%s:%s' % (
                self.commit.hexsha,
                blob.path,
              ),
            )
          else:
            contents[blob.path] = blob.data_stream.read()
      visitor = TreeVisitor(get_blob_contents)
      visitor(self.commit.tree)
      return contents

  def get_changes(self, paths: str) -> Change:
    with self.repo._lock:
      return list(change for (path, change) in self.changes.items() if path in paths)

  def get_change(self, path: str) -> Change:
    with self.repo._lock:
      return self.changes.get(path, None)

  def get_notes(self, ns: str = None) -> Notes:
    with self.repo._lock:
      return Notes(self.repo, self.commit, ns)

  def __str__(self):
    with self.repo._lock:
      s = "{} {} {} | {}".format(bracketize(self.id), self.short_message, self.author_name, self.date)
      if len(self.changes):
        s += "\n\n  "
        s += "\n  ".join(str(change) for change in self.changes.values())
        s += "\n"
      return s

  def __rich_console__(self, console, options):
    with self.repo._lock:
      yield "[bold]{}[/bold] {} | {} {}".format(bracketize(self.id), self.short_message, self.author_name, self.date)
      # for change in self.changes.values():
      #   yield change

class RevisionDetails:
  """
  Bossman creates new versions of resources through plugins. It doesn't need
  to know much about the implementation details of the plugin, but it does
  provide an opportunity to give context about an applied revision.
  
  Instances of this class join a revision id to plugin specific details, which
  could typically be a remote version number.
  """
  def __init__(self, id, details):
    self.id = id
    self.details = details

  def __str__(self):
    return "{} ({})".format(self.id, self.details)

def synchronized(method):
  def wrapper(self, *arg, **kws):
    with self._lock:
      return method(self, *arg, **kws)
  return wrapper

class Repo:

  def __init__(self, root):
    self.logger = get_class_logger(self)
    try:
      self._repo = git.Repo(root)
    except git.InvalidGitRepositoryError:
      raise RepoError("Not a git repository: {}.".format(root))
    self._lock = threading.RLock()

  def config_writer(self):
    return self._repo.config_writer()

  def config_reader(self):
    return self._repo.config_reader()

  @synchronized
  def fetch_notes(self, ns: str = "*"):
    conf = self.config_reader()
    ref = "refs/notes/" + ns
    for section in conf.sections():
      if section.startswith("remote"):
        remote = section.split(" ").pop().strip('"')
        try:
          self._repo.git.fetch(remote, "+{}:{}".format(ref, ref))
        except git.GitCommandError as e:
          if "refspec" in e.stderr.lower():
            # if we don't have notes in this namespace we might get
            # an invalid refspec or a "refspec xyz does not match any"
            #print("Warning: invalid refspec error occurred when running `{}`".format(e._cmdline), file=sys.stderr)
            pass
          elif "couldn't find remote ref" in e.stderr.lower():
            # if the remote doesn't know about this ref yet, this can happen; it is fine
            pass
          elif "authentication failures" in e.stderr:
            raise RepoAuthError("Failed to fetch from remote {}: {}".format(remote, e.stderr))
          else:
            raise e

  @synchronized
  def push_notes(self, ns: str = "*"):
    conf = self.config_reader()
    ref = "refs/notes/" + ns
    for section in conf.sections():
      if section.startswith("remote"):
        remote = section.split(" ").pop().strip('"')
        try:
          self._repo.git.push(remote, "+{}:{}".format(ref, ref))
        except git.GitCommandError as e:
          if "refspec" in e.stderr:
            # if we don't have notes in this namespace we might get
            # an invalid refspec; this should be quite unlikely
            print("Warning: invalid refspec error occurred when running `{}`".format(e._cmdline), file=sys.stderr)
            pass
          elif "does not match any" in e.stderr:
            # if we did not write any notes for this ns, we will get this error
            pass
          elif "authentication failures" in e.stderr:
            raise RepoAuthError("Failed to fetch from remote {}: {}".format(remote, e.stderr))
          else:
            raise e

  @synchronized
  def rev_parse(self, rev: str) -> str:
    # We can't use Repo.rev_parse since it doesn't support
    # short hexshas or indirect refs such as tag names or branches.
    try:
      hexsha = self._repo.git.rev_parse(rev)
      return self._repo.rev_parse(hexsha)
    except (git.GitCommandError, gitdb.exc.BadName):
      self.logger.debug("failed to resolve revision {}".format(rev))
      raise RepoRevNotFoundError("failed to resolve revision {}, please make sure it is a valid commit/tag name".format(rev))

  @synchronized
  def rev_exists(self, rev: str) -> str:
    try:
      return isinstance(self.rev_parse(rev), git.Commit)
    except BossmanError:
      return False

  @synchronized
  def rev_is_reachable(self, rev: str, from_rev: str = "HEAD") -> str:
    """
    Returns True if {rev} is an ancestor of {from_rev}.
    """
    try:
      rev = self.rev_parse(rev)
      for parent_rev in self._repo.iter_commits(from_rev):
        if parent_rev.hexsha == rev.hexsha:
          return True
      else:
        return False
    except BossmanError:
      return False

  @synchronized
  def get_paths(self, rev: str = "HEAD", predicate = true) -> list:
    """
    Lists the paths versioned for revision {rev}, optionally filtered
    by {predicate}.
    """
    commit = self.rev_parse(rev)
    tree  = commit.tree

    paths = []
    visitor = TreeVisitor(lambda blob: paths.append(blob.path))
    visitor(tree)
    return paths

  @synchronized
  def get_head(self):
    """
    Returns the HEAD revision, or None if this is the first commit.
    """
    try:
      commit = next(self._repo.iter_commits())
      prev = git.NULL_TREE
      if commit.parents:
        prev = commit.parents[0]
      diffs = commit.diff(prev, R=True)
      return Revision(self, commit, diffs)
    except StopIteration:
      return None

  @synchronized
  def get_last_revision(self, paths: list, rev: str = "HEAD") -> Revision:
    """
    Returns the last revision in {rev}'s ancestry to have affected {paths}.
    """
    try:
      rev = self.rev_parse(rev) if rev is not None else "HEAD"
      # see also the comment in self.get_revisions
      commit = next(self._repo.iter_commits(rev, paths=paths, first_parent=True))
      prev = git.NULL_TREE
      if commit.parents:
        prev = commit.parents[0]
      diffs = commit.diff(prev, paths=paths, R=True)
      return Revision(self, commit, diffs)
    except StopIteration:
      return None

  @synchronized
  def get_revision(self, rev: str, paths: list = None):
    """
    Returns the revision {rev}, with diffs for {paths}.
    """
    commit = self.rev_parse(rev)
    prev = git.NULL_TREE
    diffs = []
    if commit.parents:
      prev = commit.parents[0]
      diffs = commit.diff(prev, paths=paths, R=True) # R=True -> reverse
    return Revision(self, commit, diffs)

  @synchronized
  def get_current_user_email(self):
    """
    Returns the result of `git config user.email`
    """
    return self._repo.git.config("user.email")

  @synchronized
  def get_current_user_name(self):
    """
    Returns the result of `git config user.name`
    """
    return self._repo.git.config("user.name")


  @synchronized
  def get_current_branch(self):
    return self._repo.head.ref.name

  @synchronized
  def get_branches(self, **kwargs) -> list:
    try:
      result = self._repo.git.branch(format="%(refname:short)", **kwargs)
      return result.splitlines()
    except git.GitCommandError:
      return []

  @synchronized
  def get_branches_containing(self, rev: str, **kwargs) -> list:
    return self.get_branches(contains=rev, **kwargs)

  @synchronized
  def get_revisions(self, since_rev: str = None, until_rev: str = "HEAD",  paths: list = None) -> list:
    """
    Returns all revisions having affected {paths} in the ancestry of {until_rev}, bounded by {since_rev} if
    specified.
    """
    commitRange = ("{since_rev}..{until_rev}".format(since_rev=since_rev, until_rev=until_rev)
      if since_rev
      else until_rev)

    # Traversing history with merge support. If we have history like this:
    #
    # *  a051acd (HEAD -> main) Merge branch 'test' into main
    # |\  
    # | * 8303358 (test) cache css/jss for 30d
    # | * 96a1826 cache errors for 74s
    # |/  
    # * 4bbe5f0 (origin/main) update README
    #
    # We will traverse directly from a051acd to 4bbe5f0. For bossman, this means
    # that merge commits can effectively be used as release commits.
    commits = self._repo.iter_commits(commitRange, paths=paths, first_parent=True)
    revisions = []
    for commit in commits:
      prev = git.NULL_TREE
      if commit.parents:
        prev = commit.parents[0]
      diffs = commit.diff(prev, paths=paths, R=True) # R=True -> reverse
      # Two cases:
      # - paths were specified: we want the list of commits affecting them (with a diff)
      # - paths were not: we want the full log
      if len(diffs) or paths is None:
        revisions.append(Revision(self, commit, diffs))
    return revisions

  @synchronized
  def get_tags_pointing_at(self, rev="HEAD") -> list:
    try:
      rev = self.rev_parse(rev)
      result = self._repo.git.tag(points_at=rev)
      return result.splitlines()
    except BossmanError:
      return []


import yaml
from rich.panel import Panel
from rich.syntax import Syntax
from rich.padding import Padding

class BossmanError(RuntimeError):
  """
  Base class for bossman-specific runtime errors.
  """
  def __rich_console__(self, *args):
    error_yaml = yaml.safe_dump(self.args)
    yield Panel(
      Syntax(error_yaml, "yaml").highlight(error_yaml),
      title="[bold red]{}[/]".format(type(self).__name__),
      title_align="left",
      padding=(0, 0, 0, 4)
    )

class BossmanConfigurationError(BossmanError):
  """
  Configuration errors indicate that improper options have been passed to bossman.
  """

class MultipleMatchingPluginsError(BossmanError):
  """
  Bossman matches plugins to resources using glob patterns.
  """
  def __init__(self, resource, plugins):
    super(MultipleMatchingPluginsError, self).__init__("Only one plugin should match a resource.")
    self.resource = resource
    self.plugins = plugins

class BossmanValidationError(BossmanError):
  """
  Raised by plugins when a resource is invalid.
  """


import random
import sys, os
import time
import requests
from requests.adapters import HTTPAdapter, RetryError
from urllib3.util.retry import Retry, MaxRetryError
from akamai.edgegrid import EdgeGridAuth, EdgeRc
from configparser import NoSectionError, NoOptionError
from bossman import logging, USER_AGENT
from bossman.logging import logger
from bossman.errors import BossmanError

py3 = sys.version_info[0] >= 3
if py3:
  from configparser import ConfigParser
  from urllib import parse
else:
  import ConfigParser
  # pylint: disable=import-error
  import urlparse as parse

class EdgegridError(BossmanError):
  pass

adapter = HTTPAdapter(
  pool_connections=1,
  pool_maxsize=10,
)

class Session(logging.RequestsLoggingSession):
  def __init__(self, edgerc, section, switch_key=None, **kwargs):
    try:
      super(Session, self).__init__(**kwargs)
      self.edgerc = EdgeRc(edgerc)
      self.section = section
      self.switch_key = None
      if self.edgerc.has_option(section, "account_key"):
        self.switch_key = self.edgerc.get(section, "account_key")
      if switch_key != None:
        self.switch_key = switch_key
      self.auth = EdgeGridAuth(
        client_token=self.edgerc.get(section, "client_token"),
        client_secret=self.edgerc.get(section, "client_secret"),
        access_token=self.edgerc.get(section, "access_token"),
      )
      self.mount("https://", adapter)
      self.mount("http://", adapter)
    except NoSectionError as e:
      raise EdgegridError(e.message)
    except NoOptionError as e:
      raise EdgegridError(e.message)

  def is_retryable(self, method: str, response: requests.Response):
    if not method in ("HEAD", "GET", "PUT", "DELETE", "OPTIONS", "TRACE"):
      return False
    if 'application/problem+json' in response.headers.get('Content-Type', ''):
      if response.status_code == 400:
        problem_details = response.json()
        return problem_details.get('detail') == 'Invalid timestamp'
    return response.status_code in (429, 500, 502, 503, 504)

  def request(self, method, url, params=None, headers=None, **kwargs):
    if self.switch_key:
      params = params if params else {}
      params.update(accountSwitchKey=self.switch_key)

    headers = headers if headers else {}
    headers['user-agent'] = USER_AGENT

    baseUrl = "https://{host}".format(host=self.edgerc.get(self.section, "host"))
    url = parse.urljoin(baseUrl, url)
    # We cannot use the standard urllib3 retry mechansim because it simply retries the requests as-is,
    # with the same authorization headers. With edgegrid the auth headers may be invalid before we have
    # finished retrying, yielding "Invalid Timestamp" 400 errors.
    # https://github.com/psf/requests/issues/5975
    tries, max_tries, wait = 1, 5, 1
    while tries <= max_tries:
      response = super(Session, self).request(method, url, params=params, headers=headers, **kwargs)
      if not response.ok:
        if self.is_retryable(method, response):
          time.sleep(wait)
          wait = wait * 2 ** tries + random.uniform(0, 1)
        elif int(response.status_code) // 100 == 4:
          # input errors should be handled by the caller
          return response
        else:
          raise EdgegridError(response.json())
      else:
        return response
    raise EdgegridError('max retries exceeded')



import json
import time
import sys
from functools import lru_cache
from bossman.plugins.akamai.lib.edgegrid import Session
from bossman.logging import get_class_logger
from bossman.cache import cache
from bossman.errors import BossmanError

_cache = cache.key(__name__)

class HAPIError(BossmanError):
  pass
class HAPIValidationError(HAPIError):
  pass
class HAPIEdgeHostname(object):
  def __init__(self, chinaCdn, comments, dnsZone, edgeHostnameId, ipVersionBehavior, map, productId, recordName, securityType, serialNumber, ttl, useDefaultMap, useDefaultTtl):
    self.chinaCdn = chinaCdn
    self.comments = comments
    self.dnsZone = dnsZone
    self.edgeHostnameId = edgeHostnameId
    self.ipVersionBehavior = ipVersionBehavior
    self.map = map
    self.productId = productId
    self.recordName = recordName
    self.securityType = securityType
    self.serialNumber = serialNumber
    self.ttl = ttl
    self.useDefaultMap = useDefaultMap
    self.useDefaultTtl = useDefaultTtl

  def __str__(self):
    return self.__class__.__qualname__ + ': ' + ' '.join(f'{k}={json.dumps(v)}' for k,v in self.__dict__.items())

class HAPIClient:
  def __init__(self, edgerc, section, switch_key=None, **kwargs):
    self.logger = get_class_logger(self)
    self.session = Session(edgerc, section, switch_key=switch_key, **kwargs)

  def get_edgehostnames(self, recordNameSubstring=None, dnsZone=None):
    self.logger.debug(f'get_edgehostnames recordNameSubstring={recordNameSubstring} dnsZone={dnsZone}')
    params = dict()
    if recordNameSubstring != None:
      params['recordNameSubstring'] = recordNameSubstring
    if dnsZone != None:
      params['dnsZone'] = dnsZone
    response = self.session.get("/hapi/v1/edge-hostnames", params=params)
    edgeHostnames = []
    if response.status_code == 200:
      edgeHostnames = list(HAPIEdgeHostname(**item) for item in response.json().get("edgeHostnames", []))
    return edgeHostnames

  def get_edgehostname_by_record_name_and_zone(self, recordName, dnsZone):
    self.logger.debug(f'get_edgehostname_by_record_name_and_zone recordName={recordName} dnsZone={dnsZone}')
    edgeHostnames = list(
      edgeHostname
      for edgeHostname
      in self.get_edgehostnames(recordNameSubstring=recordName, dnsZone=dnsZone)
      if edgeHostname.recordName == recordName
      and edgeHostname.dnsZone == dnsZone
    )
    return edgeHostnames[0] if len(edgeHostnames) > 0 else None

if __name__ == '__main__':
  hc = HAPIClient('/Users/ahogg/.edgerc', 'default', switch_key='B-C-1ED34DK%3A1-8BYUX')
  ehn = hc.get_edgehostname_by_record_name_and_zone('ahogg-shared-cert-a2s2', 'akamaized.net')
  print(ehn)


import logging
import sys

import requests

import textwrap
class _Formatter(logging.Formatter):
  def formatRequestsMessage(self, record):
    _formatHeaders = lambda d: '\n'.join(f'{k}: {v}' for k, v in d.items())
    return textwrap.dedent('''
      ---------------- request ----------------
      {req.method} {req.url}
      {reqhdrs}
      {req.body}
      ---------------- response ----------------
      {res.status_code} {res.reason} {res.url}
      {reshdrs}
      {res.text}
    ''').format(
      req=record.req,
      res=record.res,
      reqhdrs=_formatHeaders(record.req.headers),
      reshdrs=_formatHeaders(record.res.headers),
    )

  def formatMessage(self, record):
    result = super().formatMessage(record)
    if record.name == 'requests':
      result += self.formatRequestsMessage(record)
    return result

_formatter = _Formatter('{asctime} {levelname} {name} {message}', style='{')

_console_handler = logging.StreamHandler()
_console_handler.setFormatter(_formatter)
_console_handler.setStream(sys.stderr)

logger = logging.getLogger()
logger.addHandler(_console_handler)

def get_class_logger(instance):
  return logger.getChild(
    "{module_name}.{class_name}".format(
      module_name=instance.__class__.__module__,
      class_name=instance.__class__.__name__
    )
  )

class RequestsLoggingSession(requests.Session):
  def __init__(self, *args, **kwargs):
    super(RequestsLoggingSession, self).__init__(*args, **kwargs)
    self.logger = logger.getChild('requests')
    self.logger.propagate = True
    self.hooks['response'].append(self.logRoundtrip)

  def logRoundtrip(self, response, *args, **kwargs):
    extra = {'req': response.request, 'res': response}
    self.logger.debug('HTTP roundtrip', extra=extra)



import parse, re, os, pathlib
from os.path import expanduser, isabs, relpath, abspath, dirname, join
from bossman.errors import BossmanConfigurationError
from bossman.logging import logger

logger = logger.getChild(__name__)

def getenv(key, default=None):
  from os import getenv
  v = getenv(key)
  if v == None:
    return default

_DEFAULT_RESOURCE_TYPES = (
  {
    "module": "bossman.plugins.akamai.property",
    "pattern": "akamai/property/{name}",
    "options": {
      "edgerc": getenv("AKAMAI_EDGERC", expanduser("~/.edgerc")),
      "section": getenv("AKAMAI_EDGERC_SECTION", "default"),
      "account_key": getenv("AKAMAI_EDGERC_SWITCHKEY", None),
    }
  },
  {
    "module": "bossman.plugins.akamai.cloudlet_v3",
    "pattern": "akamai/cloudlet/{name}",
    "options": {
      "edgerc": getenv("AKAMAI_EDGERC", expanduser("~/.edgerc")),
      "section": getenv("AKAMAI_EDGERC_SECTION", "default"),
      "account_key": getenv("AKAMAI_EDGERC_SWITCHKEY", None),
    }
  },
  {
    "module": "bossman.plugins.akamai.edgehostname",
    "pattern": "akamai/edgehostname/{name}.json",
    "options": {
      "edgerc": getenv("AKAMAI_EDGERC", expanduser("~/.edgerc")),
      "section": getenv("AKAMAI_EDGERC_SECTION", "default"),
      "account_key": getenv("AKAMAI_EDGERC_SWITCHKEY", None),
    }
  }
)

class PathPatternMatch:
  """
  canonical: the canonical resource path
  values: the variables extracted by the pattern match
  """
  def __init__(self, canonical, values):
    self.canonical = canonical
    self.values = values

class PathPattern:

  @staticmethod
  @parse.with_pattern(r'[^/]+')
  def PathComponent(s):
    return s

  def __init__(self, pattern):
    # A pattern might look like this: /a/{b}/{c}.
    # A resource path can be created using string formatting:
    #   pattern.format(b="foo", c="bar")
    self._format_pattern = pattern
    # We also need to be able to match a real path against the pattern.
    # We constrain the placeholders to only match path components, using
    # a custom regular expression. So we need to turn /a/{b}/{c} into
    # /a/{b:PathComponent}/{c:PathComponent}.
    self._match_pattern = self._format_pattern
    # So we parse the pattern with itself, retrieving dict(b="{b}", c="{c}")
    names = parse.parse(pattern, pattern).named
    # And we manipulate the match pattern based on the placeholders we discovered.
    for name, placeholder in names.items():
      placeholder_re = re.escape(placeholder)
      final_placeholder = "{{{}:PathComponent}}".format(name)
      self._match_pattern = re.sub(placeholder_re, final_placeholder, self._match_pattern)

  def match(self, path):
    """
    Matches path against the pattern. Returns a `PathPatternMatch` if successful, None otherwise.
    """
    result = parse.search(self._match_pattern, path, extra_types=dict(PathComponent=PathPattern.PathComponent))
    if result:
      canonical = self._format_pattern.format(**result.named)
      if path.startswith(canonical):
        return PathPatternMatch(canonical, result.named)
    return None

class ResourceTypeConfig:
  def __init__(self, data):
    self.pattern = PathPattern(data.get("pattern"))
    self.module = data.get("module")
    self.options = data.get("options", {})

  def __str__(self):
    return _dump(self)

class Config:
  def __init__(self, data):
    self.resource_types = (
      ResourceTypeConfig(config)
      for config
      in data.get("resources", _DEFAULT_RESOURCE_TYPES)
    )

  def __str__(self):
    return _dump(self)

def _dump(obj):
  def simplify(obj):
    try:
      __dict__ = object.__getattribute__(obj, "__dict__")
      return dict(
        (k, simplify(v))
        for (k, v)
        in __dict__.items()
      )
    except AttributeError:
      return obj
  return str(simplify(obj))

 



from collections import OrderedDict
from io import StringIO
from copy import deepcopy
import re
from bossman.repo import Revision
from bossman.errors import BossmanError


class GenericVersionComments:
  @staticmethod
  def from_revision(revision: Revision, truncate_to: int = None):
    comments = StringIO()
    comments.write(revision.message.strip())
    comments.write("\n\n")
    comments.write("commit: {}\n".format(revision.id))
    comments.write("branch: {}\n".format(", ".join(revision.branches)))
    comments.write("author: {} <{}>\n".format(revision.author_name, revision.author_email))
    if revision.author_email != revision.committer_email:
      comments.write("committer: {} <{}>\n".format(revision.commit.committer.name, revision.commit.committer.email))
    return GenericVersionComments(comments.getvalue(), truncate_to)

  def __init__(self, comments: str, truncate_to: int = None):
    self.truncate_to = truncate_to
    parts = comments.strip().rsplit("\n\n", 1)
    self.message = parts[0]
    self.metadata = (OrderedDict(re.findall(r"^([^:]+):\s*(.*)$", parts[1], re.MULTILINE))
      if len(parts) == 2
      else {})

  def __str__(self):
    def _format(msg, metadata):
      return "{}\n\n{}".format(
        msg,
        "\n".join(
          "{}: {}".format(k, v)
          for k, v in metadata.items()
        )
      )

    def _shorten(msg: str, metadata):
      """
      Yield successively shorter versions of the message.
      """
      # try the full message first
      yield _format(msg, metadata)

      # if too long, try truncating the commit message GitHub style
      msg = msg.split('\n', 1)[0][:80]
      yield _format(msg, metadata)

      # if too long, remove non-essential metadata
      for key in sorted(metadata.keys()):
        if key != 'commit':
          metadata.pop(key)
          yield _format(msg, metadata)

    for s in _shorten(self.message, deepcopy(self.metadata)):
      if self.truncate_to == None or len(s) <= self.truncate_to:
        return s

    # we should NEVER see this
    raise BossmanError("Failed to truncate the description message to the required length {}".format(self.truncate_to))

  @property
  def commit(self):
    return self.metadata.get("commit", None)

  @property
  def subject_line(self):
    # 72 chars matches GitHub guidance, but it's a bit long for the console
    return self.message.split("\n")[0][:40]

  @property
  def branch(self):
    return self.metadata.get("branch", None)

  @property
  def author(self):
    return self.metadata.get("author", None)

  @property
  def committer(self):
    return self.metadata.get("committer", None)
